<?php

namespace Drupal\quick_data\Form;

use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Url;
use Drupal\field\FieldConfigInterface;
use Drupal\paragraphs\Entity\Paragraph;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class DefaultForm.
 */
class QuickDataImportForm extends FormBase {

  /**
   * @var \Drupal\user\PrivateTempStoreFactory
   */
  protected $tempStoreFactory;

  /**
   * @var \Drupal\Core\Session\AccountInterface
   */
  private $currentUser;

  /**
   * @var \Drupal\user\PrivateTempStore
   */
  protected $store;

  /**
   * Constructs a \Drupal\demo\Form\Multistep\MultistepFormBase.
   *
   * @param \Drupal\Core\TempStore\PrivateTempStoreFactory $temp_store_factory
   * @param \Drupal\Core\Session\AccountInterface $current_user
   */

  protected $step = 1;

  protected string $separate = " ; ";

  protected array $fieldTypeException = [
    'comment',
    'feeds_item',
    'tablefield',
    'entity_reference_revisions',
  ];

  public function __construct(PrivateTempStoreFactory $temp_store_factory, AccountInterface $current_user) {
    $this->tempStoreFactory = $temp_store_factory;
    $this->currentUser = $current_user;
    $this->store = $this->tempStoreFactory->get('quickDataImport');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('tempstore.private'),
      $container->get('current_user')
    );
  }

  public function buildForm(array $form, FormStateInterface $form_state, $type = '', $bundle = '', $field = '', $entity_id = '') {
    $form_state->set('type', $type);
    $form_state->set('bundle', $bundle);
    $form_state->set('field', $field);
    $form_state->set('entity_id', $entity_id);

    if ($this->step != 1) {
      return $this->quick_data_import_process_csv($form, $form_state);
    }
    return $this->quick_data_import_load_csv($form, $form_state);
  }

  public function getFormId() {
    return 'quick_data_import_form';
  }

  public function submitForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    if (!empty($values["csv_file_upload"])) {
      $form_state->set('csv_file_upload', $form_state->getValue('csv_file_upload'));
      $form_state->set('csv_delimiter', $form_state->getValue('csv_delimiter'));
      $form_state->setRebuild();
      $this->step = 2;
    }
    if (!empty($values["import"])) {
      $store = $form_state->getStorage();
      $num_imported = $this->quick_data_import_mapping($store, $form_state->getUserInput()['col']);
      if ($num_imported) {
        \Drupal::messenger()
          ->addStatus($this->t('@count imported items total.', ['@count' => $num_imported]));
      }
      // redirection! This is because  $form_state['rebuild'] gets set to TRUE
      if (!empty($_GET['destination'])) {
        $dest_url = $_GET['destination'];
        $parse = parse_url($dest_url);
        if(!empty($parse['host'])){
          $response = new TrustedRedirectResponse(Url::fromUri($dest_url)->toString());
          $metadata = $response->getCacheableMetadata();
          $metadata->setCacheMaxAge(0);
          $form_state->setResponse($response);
        }else{
          $url = Url::fromUri('internal:/' . $dest_url);
        }
      }
      elseif (!empty($store['entity_id'])) {
        $dest_url = implode('/', [$store['bundle'], $store['entity_id']]);
        $url = Url::fromUri('internal:/' . $dest_url);
      }
      if (!empty($url)) {
        $form_state->setRedirectUrl($url);
      }
    }

  }

  public function quick_data_import_load_csv($form, &$form_state) {
    $type = $form_state->get('type');
    $bundle = $form_state->get('bundle');
    $field = $form_state->get('field');
    $entity_id = $form_state->get('entity_id');
    //page one load csv
    if (!in_array($type, [
      "node",
      "user",
      "taxonomy_term",
      "data",
      "paragraph",
      "field_collection_item",
    ])) {
      \Drupal::messenger()
        ->addError(t("Your type %type is not available", ['%type' => $bundle]));
      return $form;
    }
    //check bundle validate
    $filename = implode('-', [$type, $bundle]) . '.csv';
    if (in_array($type, ['field_collection_item', 'paragraph'])) {
      if (empty($field) || empty($bundle) || empty($entity_id)) {
        \Drupal::messenger()->addError(t("Yours variables is not available"));
        return $form;
      }
      $filename = implode('-', [$bundle, $field]) . '.csv';
    }

    $fields_info = $this->_fields_info($bundle, $type, $field, $entity_id);
    $options_fields = $this->_options_load_field($fields_info);
    $remove_none = array_shift($options_fields);
    $form['files']['csv_file_upload'] = [
      '#type' => 'managed_file',
      '#title' => t('Choose a file'),
      '#description' => '<a id="export" data-filename="' . $filename . '" download="' . $filename . '"> ' . t("Download CSV template: ") . '</a>
<span id="export-header" >' . implode(';', $options_fields) . '</span>
<br/><i>' . t('Use "%separate" for separate multiple values', ['%separate' => $this->separate]) . '</i>
',
      '#required' => TRUE,
      '#upload_validators' => [
        'file_validate_extensions' => ['csv'],
      ],
    ];
    $form['files']['csv_delimiter'] = [
      '#type' => 'select',
      '#title' => t('Character used to separate'),
      '#options' => [
        ';' => ';',
        ',' => ',',
      ],
    ];

    $form['#attached'] = [
      'library' => ['quick_data/drupal.quick_data'],
      'drupalSettings' => [
        'quick_data' => [
          'entity_type' => $form_state->getValue('type'),
        ],
      ],
    ];
    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => t('Next'),
      '#button_type' => 'primary',
    ];
    return $form;
  }

  protected function quick_data_import_process_csv($form, &$form_state) {
    $type = $form_state->get('type');
    $bundle = $form_state->get('bundle');
    $field = $form_state->get('field');
    $entity_id = $form_state->get('entity_id');
    $fields_info = $this->_fields_info($bundle, $type, $field, $entity_id);
    $form_state->set('fields_info', $fields_info);
    $options_fields = $this->_options_load_field($fields_info);
    $field = [
      '#type' => 'select',
      '#options' => $options_fields,
      '#empty_option' => t('- Select -'),
    ];
    $flip_fields = !empty($options_fields) ? array_flip($options_fields) : [];
    $csv_rows = $this->_convert_csv_array($form_state->getValue("csv_file_upload"), $form_state->getValue("csv_delimiter"), 10);
    $header = [];
    $header_title = array_shift($csv_rows);
    foreach ($header_title as $col => $title_col) {
      $title_col = trim($title_col);
      $header[$col]["data"] = $field;
      $header[$col]["data"]["#title"] = $title_col;
      $header[$col]["data"]["#name"] = 'col[' . $col . ']';
      if (!empty($flip_fields[$title_col])) {
        $header[$col]["data"]["#value"] = $flip_fields[$title_col];
      }
    }

    $form['table'] = [
      '#theme' => 'table',
      "#title" => t("Mapping field with data"),
      '#header' => $header,
      '#rows' => $csv_rows,
      '#empty' => t('Empty'),
    ];

    $form['actions']['#type'] = 'actions';
    $form['actions']['import'] = [
      '#type' => 'submit',
      '#value' => t('Import'),
      '#button_type' => 'primary',
    ];
    return $form;
  }

  public function _convert_csv_array($fid, $delimiter = ';', $limit = FALSE) {
    if (is_array($fid)) {
      $fid = end($fid);
    }
    $file = \Drupal\file\Entity\File::load($fid);
    $filename = $file->get('uri')->value;
    $csv_rows = [];
    if (($f = fopen($filename, "r")) !== FALSE) {
      $i = 0;
      while (($data = fgetcsv($f, 1000, $delimiter)) !== FALSE) {
        $csv_rows[] = $data;
        if ($limit && $i++ > $limit) {
          break;
        }
      }
      fclose($f);
    }
    return $csv_rows;
  }

  private function _table_info($table) {
    $schema = drupal_get_module_schema('quick_data', $table);
    $infos = [];
    $i = 0;
    foreach ($schema['fields'] as $name => $field) {
      $infos[$name] = [
        'field_name' => $name,
        'label' => !empty($field['description']) ? $field['description'] : $name,
        'type' => $field["type"],
        'columns' => ['value' => ['type' => $field["type"]]],
        'cardinality' => 1,
        'weight' => $i++,
      ];
    }
    return $infos;
  }

  protected function _get_target($field_definition) {
    $propertyDefinitions = $field_definition->getPropertyDefinitions();
    $module = $field_definition->get('module');
    $type = $field_definition->getType();
    $targetDefinition = [
      'text_with_summary' => ['value', 'summary'],
      'link' => ['uri', 'title'],
      'address' => [
        'organization',
        'given_name',
        'family_name',
        'country_code',
        'address_line1',
        'address_line2',
        'postal_code',
        'locality',
        //'langcode',
        //'administrative_area',
        //'dependent_locality',
        //'additional_name',
        //'sorting_code',
      ],
      'name' => [
        'title',
        'given',
        'middle',
        'family',
        //'generational',
        //'credentials',
      ],
      'daterange' => ['value', 'end_value'],
      'entity_reference' => ['target_id'],
      'image' => ['target_id', 'alt', 'title'],
      'double_field' => ['first', 'second'],
    ];
    $target = FALSE;
    if (!empty($targetDefinition[$type])) {
      foreach ($targetDefinition[$type] as $column) {
        $target[$column] = (string) $propertyDefinitions[$column]->getLabel();
      }
    }
    return $target;
  }

  private function _fields_info($bundle, $type, $field = '', $entity_id = FALSE) {
    if ($bundle == 'data') {
      return $this->_table_info($type);
    }
    if ($type == 'paragraph') {
      $host = \Drupal::entityTypeManager()
        ->getStorage($bundle)
        ->load($entity_id);
      $content_type = $host->getType();
      $field_instance = \Drupal::entityTypeManager()
        ->getStorage('field_config')
        ->load(implode('.', [$bundle, $content_type, $field]));
      $targetBundle = array_key_first($field_instance->getSetting("handler_settings")["target_bundles"]);
      /** @var \Drupal\paragraphs\ParagraphInterface $paragraphs_entity */
      $paragraphs_entity = \Drupal::entityTypeManager()->getStorage($type)
        ->create(['type' => $targetBundle]);
      $field_definitions = $paragraphs_entity->getFieldDefinitions();
    }
    else {
      $field_definitions = \Drupal::service('entity_field.manager')
        ->getFieldDefinitions($type, $bundle);
    }

    foreach ($field_definitions as $name => $definition) {
      if ($field_definitions[$name] instanceof FieldConfigInterface) {
        if (empty($field_type) || $definition->getType() == $field_type) {
          $fields[$name] = $definition;
        }
      }
    }
    $infos = [];
    $repository = \Drupal::service('entity_display.repository');
    $formDisplay = $repository->getFormDisplay($type, $bundle, 'default');
    if (!empty($fields)) {
      foreach ($fields as $field) {
        $field_type = $field->getType();
        $field_info = $field->getFieldStorageDefinition();
        $module = $field_info->get('module');
        if (in_array($field_type, $this->fieldTypeException)) {
          continue;
        }
        $name = $field->getName();
        $fieldComponent = $formDisplay->getComponent($name);
        $infos[$name] = [
          'field_name' => $name,
          'label' => $field->getLabel(),
          'type' => $field->getType(),
          'widget' => $field_info->getType(),
          'columns' => $this->_get_target($field_info),
          'cardinality' => $field_info->getCardinality(),
          'allowed_values' => options_allowed_values($field_info),
          'weight' => isset($fieldComponent["weight"]) ? $fieldComponent["weight"] : 0,
          'module' => $module,
          'schema' => $field_info->getSchema()['columns'],
          'settings' => $field->getSettings(),
        ];
      }
    }
    switch ($type) {
      case 'node':
        $infos['status'] = [
          'field_name' => 'status',
          'label' => t('Status'),
          'type' => 'status',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['title'] = [
          'field_name' => 'title',
          'label' => t('Title'),
          'type' => 'title',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;
      case 'user':
        $infos['status'] = [
          'field_name' => 'status',
          'label' => t('Status'),
          'type' => 'status',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['mail'] = [
          'field_name' => 'mail',
          'label' => t('Email '),
          'type' => 'mail',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['name'] = [
          'field_name' => 'name',
          'label' => t('Name'),
          'type' => 'name',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;
      case 'taxonomy_term':
        $infos['description'] = [
          'field_name' => 'description',
          'label' => t('description'),
          'type' => 'description',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['name'] = [
          'field_name' => 'name',
          'label' => t('Name'),
          'type' => 'name',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;
    }

    return $infos;
  }

  private function _options_load_field($infos) {
    uasort($infos, 'Drupal\Component\Utility\SortArray::sortByWeightElement');
    $options = ['_none' => (string) t('- None -')];
    foreach ($infos as $field_name => $info) {
      if (in_array($info['type'], $this->fieldTypeException)) {
        continue;
      }
      if (!empty($info['columns'])) {
        foreach ($info['columns'] as $name => $label) {
          $options["$field_name:$name"] = trim((string) $info['label'] . ': ' . $label);
        }
        continue;
      }
      $options[$field_name] = (string) $info['label'];
    }
    return $options;
  }

  protected function quick_data_import_mapping($store, $mapping) {
    $data = $this->_convert_csv_array($store['csv_file_upload'], $store['csv_delimiter']);
    $header_title = array_shift($data);
    $id = $store['entity_id'];
    $host = FALSE;
    $num_imported = 0;
    if (in_array($store['type'], [
        'field_collection_item',
        'paragraph',
      ]) && $id) {
      switch ($store['bundle']) {
        case 'node':
          $host = \Drupal\node\Entity\Node::load($id);
          break;
        case 'user':
          $host = \Drupal\user\Entity\User::load($id);
          break;
        case 'field_collection_item':
          $host = \Drupal\field_collection\Entity\FieldCollection::load($id);
          break;
        case 'paragraph':
          $host = \Drupal\paragraphs\Entity\Paragraph::load($id);
          break;
        case 'taxonomy_term':
          $host = \Drupal\taxonomy\Entity\Term::load($id);
          break;
        default:
          $entity_storage = \Drupal::entityTypeManager()
            ->getStorage($store['bundle']);
          $host = $entity_storage->load($id);
          break;
      }
    }

    foreach ($data as $row_index => $row) {
      $field_values = [];
      foreach ($row as $col => $value) {
        if (empty($value) || $mapping[$col] == '_none') {
          continue;
        }
        $field = $mapping[$col];
        $checkTarget = explode(':', $field);
        $target = FALSE;
        if (!empty($checkTarget[1])) {
          $field = $checkTarget[0];
          $target = $checkTarget[1];
        }
        $table_field[$field] = $value;
        $this->quick_data_prepare_data($field_values, $store['fields_info'][$field], $value, $target);
      }
      if (empty($field_values)) {
        return FALSE;
      }
      switch ($store['type']) {
        case 'node':
          $default = [
            'type' => $store['bundle'],
            'uid' => \Drupal::currentUser()->id(),
            'status' => 1,
          ];
          $entity = \Drupal\node\Entity\Node::create(array_merge($default, $field_values));
          try {
            $entity->enforceIsNew();
            $entity->save();
          } catch (\Exception $e) {
            \Drupal::messenger()->addError($e->getMessage());
          }
          break;
        case 'user':
          $entity = \Drupal\user\Entity\User::create($field_values);
          $entity->enforceIsNew();
          $entity->activate();
          $entity->setPassword(user_password(6));
          $entity->save();
          break;
        case 'taxonomy_term':
          $default = ['vid' => $store['bundle']];
          $entity = \Drupal\taxonomy\Entity\Term::create(array_merge($default, $field_values));
          $entity->save();
          break;
        case 'paragraph':
          if ($host) {
            $content_type = $host->getType();
            $field_instance = \Drupal::entityTypeManager()
              ->getStorage('field_config')
              ->load(implode('.', [
                $host->getEntityTypeId(),
                $content_type,
                $store['field'],
              ]));
            $paragraph_type = array_key_first($field_instance->getSetting("handler_settings")["target_bundles"]);
            $default = [
              'type' => $paragraph_type,
              'parent_id' => $host->id(),
              'parent_type' => $content_type,
              'parent_field_name' => $store['field'],
            ];
            $paragraph = \Drupal\paragraphs\Entity\Paragraph::create(array_merge($default, $field_values));
            $paragraph->isNew();
            $paragraph->save();
            $entity_items[] = [
              'target_id' => $paragraph->id(),
              'target_revision_id' => $paragraph->getRevisionId(),
            ];
          }
          break;
        case 'field_collection_item':
          if ($host) {
            //todo
            $content_type = $host->getType();
            $field_instance = \Drupal::entityTypeManager()
              ->getStorage('field_config')
              ->load(implode('.', [
                $host->getEntityTypeId(),
                $content_type,
                $store['field'],
              ]));

          }
          break;
        case 'data':
          if (!empty($table_field)) {
            $nid = db_insert($store['bundle'])
              ->fields($table_field)
              ->execute();
          }
          break;
        default:
          $entity = \Drupal::entityTypeManager()
            ->getStorage($store['type'])
            ->create($field_values);
          $entity->save();
          break;
      }
      $num_imported++;
    }
    if (!empty($entity_items)) {
      foreach ($entity_items as $entity_item) {
        $host->get($store['field'])->appendItem($entity_item);
      }
      $host->save();
    }
    return $num_imported;
  }

  protected function quick_data_prepare_data(&$field_values, $field_info, $value, $target = FALSE) {
    $field = $field_info['field_name'];
    $timeFormat = 'Y-m-d H:i:s';
    if (!empty($field_info["columns"]) && !$target) {
      $target = array_key_first($field_info["columns"]);
      $type = $field_info["columns"][$target];
    }
    if ($field_info['cardinality'] == -1) {
      $values = explode($this->separate, $value);
    }
    else {
      $values = [$value];
    }
    foreach ($values as $index => $value) {
      switch ($field_info['type']) {
        case 'taxonomy_term_reference':
          if (!is_numeric($value)) {
            $term = \Drupal::entityTypeManager()
              ->getStorage('taxonomy_term')
              ->loadByProperties(['name' => $value]);
            //create term if it is not exist
            if (!$term) {
              $term = \Drupal\taxonomy\Entity\Term::create([
                'name' => $value,
                'vid' => $field_info["allowed_values"][0]["vocabulary"],
              ]);
              $term->save();
            }
            if (!empty($term->tid)) {
              $value = $term->tid;
            }
          }
          break;
        case 'user_reference':
          if (!is_numeric($value)) {
            $user_load = user_load_by_name($value);
            if ($user_load) {
              $value = $user_load->uid;
            }
          }
          break;

        case 'node_reference':
        case 'entity_reference':
          if (!is_numeric($value)) {
            $nodes = \Drupal::entityTypeManager()
              ->getStorage('node')
              ->loadByProperties(['title' => trim($value)]);
            if ($nodes) {
              foreach ($nodes as $node) {
                if (!empty($field_info["settings"]["handler_settings"]["target_bundles"][$node->getType()])) {
                  $value = $node->id();
                }
              }
            }
          }
          break;

        case 'list_integer':
        case 'list_float':
        case 'list_boolean':
        case 'list_text':
          if (empty($field_info["allowed_values"][$value])) {
            $flip_allowed_values = array_flip($field_info["allowed_values"]);
            $value = !empty($flip_allowed_values[$value]) ? $flip_allowed_values[$value] : NULL;
          }
          break;

        case 'number_integer':
        case 'number_float':
        case 'number_decimal':
          if (!is_numeric($value)) {
            $value = NULL;
          }
          break;
        case 'file':
        case 'image':
          if (!is_integer($value)) {
            if (\Drupal\Component\Utility\UrlHelper::isValid($value, TRUE)) {
              $arr = parse_url($value);
              if (!empty($arr)) {
                $paths = explode('/', $arr['path']);
                $filename = end($paths);
                $data = file_get_contents($value);
                if (!empty($field_info["settings"]["file_directory"])) {
                  $token = \Drupal::token();
                  $field_info["settings"]["file_directory"] = $token->replace($field_info["settings"]["file_directory"]);
                  $directory = 'public://' . $field_info["settings"]["file_directory"];
                  \Drupal::service('file_system')
                    ->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY);
                  $filename = $field_info["settings"]["file_directory"] . '/' . $filename;
                }
                $file = file_save_data($data, 'public://' . $filename);
                $value = $file->id();
              }
            }
            else {
              $files = \Drupal::entityTypeManager()
                ->getStorage('file')
                ->loadByProperties(['uri' => $value]);
              if (!empty($files)) {
                $value = array_keys($files)[0];
              }
            }
          }
          break;

        case 'date' :
        case 'datetime' :
        case 'daterange' :
        case 'datestamp' :
          $timeConvert = date($timeFormat, strtotime(str_replace('/', '-', $value)));
          $dateTime = \DateTime::createFromFormat($timeFormat, $timeConvert);
          $value = $dateTime->format('Y-m-d\TH:i:s');
          if ($field_info['type'] == 'datestamp') {
            $value = $dateTime->getTimestamp();
          }
          break;
      }
      if (!is_null($value)) {
        if (!$target) {
          $target = 'value';
        }
        if (!empty($field_info['schema'][$target]['length'])) {
          $value = mb_substr($value, 0, $field_info['schema'][$target]['length']);
        }
        $field_values[$field][$index][$target] = $value;
      }
    }
  }
}
